from __future__ import with_statement
print "b56"
os.chdir("/home/rlmill/sage/aly_notebook/worksheets/admin/1/cells/4")
def LLL_toy(Basis_Matrix):

#Warning: this only consistently works with Basis_Matrix of size 2x2.
#With size 3x3 or larger, the bottom row is often off by a multiple of the other two rows
#I'm not sure why yet.

#This algorithm is the basic LLL algorithm found on page 88 of Cohen's Computational
#Number Theory text, specifically algorithm 2.6.3

#Background
# we will be using the typical Gram-Schmidt notation for the Gram-Schmidt process
# with meu for the greek letter mue and basis_star for b*
# a basis is called LLL reduced if
#     1. abs(meu[i,j])<=1/2 for 0<=j<i<=n
# and 2. basis_star[i]*basis_star[i]>=(3/4-meu[i,i-1]^2)*basis_star[i-1]*basis_star[i-1]

#Important points about LLL
#  This algorithm only computes Gram-Schmidt coefficients meu[k,j] and B[k]
#  when needed, thus step 2 is just going through one iteration of the Gram-Schmidt process
#  Which have been computed is recorded in the value k_max, which is the larges k-value attained

#note: the following are some assumption which are made when the algorithm is run

# First: that the input matrix does in fact represent a basis, this is first checked
#  in the very first if statement and makes the B[k]=0 redundant
#  Cohen states that the assumption of linearly independent row vectors is can be avoided
#  and is in fact in his algorithm 2.6.8 (Cohen: Computational Number Theory)

# Second: the matrix entries are rational, this is just for ease of my own checking and
#   for greater precision and could be replaced once the algorithm is consistently working

#check if this is in fact a basis because the algorithm does not work if it isn't

    if Basis_Matrix.rank() != Basis_Matrix.nrows():
        raise NotImplementedError,"Not a Basis"

#step 1
# in this step the matrices meu,B,basis_star (b* in Gram-Schmidt notation), and H are initialized
# meu,basis_star, and B are all zeros and H is the identity matrix
# the first Gram-Schmidt values of basis_star and B are then computed (i.e. for the 0th row)

    basis_star,meu,k,k_max,B,H = Initialize(Basis_Matrix)
    n = Basis_Matrix.nrows()
    Lovazs = Integer(0)

#Mp collects the matrix objects used to graphically compare the values of meu
#BM is the collection of b[i]*b[i] to compare how this converges
    Bl = [list() for i in range(n)]
    BMl =[list() for i in range(n)]
    Mp = [matrix_plot(meu)]  #for animation of meu values as they converge
    for i in range(Basis_Matrix.nrows()):
        Bl[i] = [line([(i/n,Integer(0)),(i/n,B[i])])]
        if i == Integer(0):
            Lovazs = Integer(0)
        else:
            Lovazs = (Integer(3)/Integer(4) - meu[i,i-Integer(1)]**Integer(2))*B[i-Integer(1)]
        BMl[i] = [line([((i+Integer(1)/Integer(2))/n,Integer(0)),((i+Integer(1)/Integer(2))/n,Lovazs)],rgbcolor=(Integer(1),Integer(0),Integer(0)))]

#step 2
# this is just one step of the Gram-Schmidt process, i.e. here we are computing only what we need
# to check whether the basis is or is not reduced
# we use k_max to keep track of where we are in the Gram-Schmidt process, keeping k_max = to the maximum
# value of k which we have currently obtained

    for k in range(Integer(1),n):

        if k > k_max:
            k_max = k
            basis_star = swap_rows(k,k,basis_star,Basis_Matrix) #sets the k'th row of basis_star = k'th of Basis_Matrix

            for j in range(k):
                meu[k,j] = Basis_Matrix[k]*basis_star[j]/B[j]
                basis_star = basis_star.rows();
                basis_star[k] = basis_star[k]-meu[k,j]*basis_star[j]
                basis_star = matrix(basis_star)


            B[k] = basis_star[k]*basis_star[k]

            if B[k]==RealNumber('0.0'):
                return('error Bk = 0, not a basis')
        Mp.append(matrix_plot(meu))
        for i in range(n):
            Bl[i].append(line([(i/n,Integer(0)),(i/n,B[i])]))
            if i == Integer(0):
                Lovaz = Integer(0)
            else:
                Lovazs = (Integer(3)/Integer(4) - meu[i,i-Integer(1)]**Integer(2))*B[i-Integer(1)]
            BMl[i].append(line([((i+Integer(1)/Integer(2))/n,Integer(0)),((i+Integer(1)/Integer(2))/n,Lovazs)],rgbcolor=(Integer(1),Integer(0),Integer(0))))


#step 3 [This is where all the work is done]
# in this step we continually run the RED subroutine
# the RED subroutine checks if meu[k,l] is reduced, i.e. less than or equal to 1/2 and if not
# it reduces meu[k,l] by q = floor(meu+1/2) or the closest integer to meu[k,l] ensuring that
# meu[k,l] is then less than or equal to 1/2
# now notice two things, first, as we are working inductively, we can assume that for l<j<k, meu[k,j] <= 1/2
# and basis_star[j] is orthogonal to Basis_Matrix[l].  Second, replacing Basis_Matrix[k] by
# Basis_Matrix[k]-q*Basis_Matrix[l] is equivalent to replacing meu[k,l] by meu[k,l]-q,
# and for j>l, meu[k,j] is not modified, finally, for 0<=i<l, the algorithm updates meu[k,i] = meu[k,i]-q*meu[l,i]
# so that the matrix meu is consistent with the matrices Basis_Matrix and basis_star as in the Gram-Schmidt process

# Next we check the other condition of LLL reduced, (the Lovasz condition) B[k]>=(3/4-meu[k,k-1]^2)*B[k-1]
# if this condition is satisfied we increase k by one and continue on, otherwise we need to use the SWAP algorithm
# The SWAP algorithm swaps Basis_Matrix[k] and Basis_Matrix[k-1], updates meu, basis_star, B, and H and then decreases
# k by one (unless k = 1, then k is reset to k) for the following iteration of step 3 where both conditions are again
# checked.

        meu,Basis_Matrix,H = RED(k,k-Integer(1),meu,Basis_Matrix,H)
        while B[k]<(RealNumber('0.75')-meu[k,k-Integer(1)]**Integer(2))*B[k-Integer(1)]:
#        while B[k]<(0.75-meu[k,k-1]^2)*B[k-1]:
            meu,B,basis_star,Basis_Matrix,H = SWAP(k,k_max,meu,B,Basis_Matrix,basis_star,H)
            k = max(Integer(1),k-Integer(1))
            Mp.append(matrix_plot(meu))
            for i in range(n):
                Bl[i].append(line([(i/n,Integer(0)),(i/n,B[i])]))
                if i == Integer(0):
                    Lovaz = Integer(0)
                else:
                    Lovazs = (Integer(3)/Integer(4) - meu[i,i-Integer(1)]**Integer(2))*B[i-Integer(1)]
                BMl[i].append(line([((i+Integer(1)/Integer(2))/n,Integer(0)),((i+Integer(1)/Integer(2))/n,Lovazs)],rgbcolor=(Integer(1),Integer(0),Integer(0))))


            meu,Basis_Matrix,H = RED(k,k-Integer(1),meu,Basis_Matrix,H)

        if B[k] >= (RealNumber('0.75')-meu[k,k-Integer(1)]**Integer(2))*B[k-Integer(1)]:
            for i in range(Integer(2),k-Integer(1)):
                l = k-i
                meu,Basis_Matrix,H = RED(k,k-Integer(1),meu,Basis_Matrix,H)
            Mp.append(matrix_plot(meu))
            for i in range(n):
                Bl[i].append(line([(i/n,Integer(0)),(i/n,B[i])]))
                if i == Integer(0):
                    Lovaz = Integer(0)
                else:
                    Lovazs = (Integer(3)/Integer(4) - meu[i,i-Integer(1)]**Integer(2))*B[i-Integer(1)]
                BMl[i].append(line([((i+Integer(1)/Integer(2))/n,Integer(0)),((i+Integer(1)/Integer(2))/n,Lovazs)],rgbcolor=(Integer(1),Integer(0),Integer(0))))


    print(Basis_Matrix)
    print(B)
#    a = animate(Mp)
#    a.show()
    g.plot((Integer(0),Integer(0)))
    for i in range(n):
        c = animate(Bl[i])
        d = animate(BMl[i])
        g = g+c+d
    g.show()
    return(Basis_Matrix)

print "e56"
